#include <windows.h>
#include <mmsystem.h>
#if defined __WIN32__ || defined _WIN32 || defined __NT__
  #include <commctrl.h>
#endif
#include "..\include\wavemix.h"
#include "mixcfg.h"

static HINSTANCE hinst;
static HWND hwndDialog;
static FARPROC lpfnTestPadPrev;
static HANDLE hMixSession;
static LPMIXWAVE lpMix1;  /* left button */
static LPMIXWAVE lpMix2;  /* right button */
static LPMIXWAVE lpMix3;  /* continuous sound */
static BOOL PlayBackgr=FALSE;

static char *szInfo[] = { "Click in the test pad to test the current settings.\n\n"
                          "Adjust the settings if needed. Click on the options "
                          "(at the left) to get more detailed information.",
                          // -----
                          "\"Remix\" has low latency, but on some sound cards "
                          "the \"reset operation\" causes an audible \"click\".\n\n"
                          "Use \"queue\" or \"remix without reset\" if your "
                          "sound card produces \"clicks\".",
                          #if 0
                            "Low latency means that the period between a click "
                            "of a mouse button on the test pad and the sound "
                            "effect coming out of the speaker is short.",
                          #endif
                          // -----
                          "\"Queue\" stores any new buffer at the end of the "
                          "list of buffers that the sound card is currently "
                          "playing.\n\n"
                          "This means that latency is high, especially if "
                          "WaveMix uses many buffers (see the \"Buffers\" "
                          "option).",
                          // -----
                          "\"Remix without reset\" uses a different algorithm "
                          "to remix the buffers, in order to avoid an audible "
                          "click while keeping low latency.\n\n"
                          "Not all sound cards support the \"remix without reset\" "
                          "option, however.",
                          // -----
                          "A sampling frequency of 11025\xa0Hz has less risk of "
                          "running out of buffers, at the cost of a lower "
                          "audio quality.\n\n"
                          "This is the default sampling rate of WaveMix.",
                          // -----
                          "A sampling frequency of 22050\xa0Hz gives a higher "
                          "audio quality, but you may need more buffers to "
                          "avoid small gaps in the sound.\n\n"
                          "See the \"Buffers\" option on how to adjust the "
                          "number of buffers.",
                          // -----
                          "If you can use a \"remix\" mix method (with or "
                          "without reset), choose a high number of buffers.\n\n"
                          "When using the \"queue\" mix method, however, use "
                          "the lowest setting that outputs the audio without gaps."
                        };

#if defined __WIN32__ || defined _WIN32 || defined __NT__
#else
HINSTANCE hinst3D;              /* instance handle of DLL for 3D effects */
static BOOL (WINAPI *Ctl3dRegister)(HINSTANCE);
static BOOL (WINAPI *Ctl3dUnregister)(HINSTANCE);
static BOOL (WINAPI *Ctl3dAutoSubclass)(HINSTANCE);
#endif

static void CentreWindow(HWND hwnd)
{
  RECT rect;
  int width,height;

  GetWindowRect(hwnd,&rect);
  width=GetSystemMetrics(SM_CXSCREEN);
  height=GetSystemMetrics(SM_CYSCREEN);
  SetWindowPos(hwnd,0,
               (width - (rect.right-rect.left))/2,
               (height - (rect.bottom-rect.top))/2,
               0,0,SWP_NOSIZE|SWP_NOACTIVATE|SWP_NOZORDER);
}

static void CloseWaveMix(void)
{
  if (hMixSession==NULL)
    return;

  WaveMixCloseChannel(hMixSession,0,WMIX_ALL);

  if (lpMix1!=NULL)
    WaveMixFreeWave(hMixSession,lpMix1);
  if (lpMix2!=NULL)
    WaveMixFreeWave(hMixSession,lpMix2);
  if (lpMix3!=NULL)
    WaveMixFreeWave(hMixSession,lpMix3);

  WaveMixCloseSession(hMixSession);

  lpMix1=lpMix2=lpMix3=NULL;
  hMixSession=NULL;
}

static void PlayWave(int channel,LPMIXWAVE lpWave,BOOL loop)
{
  MIXPLAYPARAMS params;

  params.wSize=sizeof(MIXPLAYPARAMS);
  params.hMixSession=hMixSession;
  params.iChannel=channel;
  params.lpMixWave=lpWave;
  params.hWndNotify=NULL;
  params.dwFlags=WMIX_CLEARQUEUE|WMIX_HIPRIORITY;
  params.wLoops= loop ? 0xffff : 0;

  WaveMixPlay(&params);
}

LRESULT CALLBACK TestPadShell(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{
  RECT rect;
  int x,y;

  switch (message) {
  case WM_NCHITTEST:
    GetWindowRect(hwnd,&rect);
    x=LOWORD(lParam);
    y=HIWORD(lParam);
    if (x>rect.left && x<rect.right && y>rect.top && y<rect.bottom) {
      if (hMixSession==NULL) {
        hMixSession=WaveMixInit();
        WaveMixOpenChannel(hMixSession,0,WMIX_OPENSINGLE);
        lpMix1=WaveMixOpenWave(hMixSession,"wave1",hinst,WMIX_RESOURCE);
        WaveMixOpenChannel(hMixSession,1,WMIX_OPENSINGLE);
        lpMix2=WaveMixOpenWave(hMixSession,"wave2",hinst,WMIX_RESOURCE);
        WaveMixOpenChannel(hMixSession,2,WMIX_OPENSINGLE);
        lpMix3=WaveMixOpenWave(hMixSession,"wave3",hinst,WMIX_RESOURCE);
        WaveMixActivate(hMixSession,TRUE);
        if (PlayBackgr)
          PlayWave(2,lpMix3,TRUE);
        SetCapture(hwnd);
      } /* if */
    } else {
      ReleaseCapture();
      CloseWaveMix();
    } /* if */
    break;
  case WM_MOUSEMOVE:
    GetClientRect(hwnd,&rect);
    x=LOWORD(lParam);
    y=HIWORD(lParam);
    if (x<0 || x>=rect.right || y<0 || y>=rect.bottom) {
      ReleaseCapture();
      CloseWaveMix();
    } /* if */
    break;
  case WM_LBUTTONDOWN:
  case WM_LBUTTONDBLCLK:
    PlayWave(0,lpMix1,FALSE);
    break;
  case WM_RBUTTONDOWN:
  case WM_RBUTTONDBLCLK:
    PlayWave(1,lpMix2,FALSE);
    break;
  } // switch
  return CallWindowProc(lpfnTestPadPrev,hwnd,message,wParam,lParam);
}


#pragma argsused
BOOL CALLBACK MixConfDlg(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
static HFONT hfont;
static int dlgcX,dlgcY;
static HBRUSH hbrushGreen;
static int orgMix,orgFreq,orgBuffers;
  RECT rect;
  int curMix,curFreq,curBuffers;
  int pos;

  switch (msg) {
  case WM_INITDIALOG:
    hwndDialog=hwnd;
    // Use MapDialogRect() instead of GetDialogBaseUnits(), because the base
    // units are always related to the system font. A dialog may or may not
    // use the system font. In fact, the dialogs of COMMDLG.DLL use MSSANS
    // 8 pt instead of the system font.
    SetRect(&rect,0,0,4,8);
    MapDialogRect(hwnd,&rect);
    dlgcX=rect.right;
    dlgcY=rect.bottom;
    // read current settings, open WaveMix just for reading the settings, then close it
    hMixSession=WaveMixInit();
    orgMix=WaveMixGetProfileInt("Remix",1);
    orgFreq=WaveMixGetProfileInt("SamplesPerSec",11);
    orgBuffers=WaveMixGetProfileInt("WaveBlocks",3);
    WaveMixCloseSession(hMixSession);
    hMixSession=NULL;
    CheckDlgButton(hwnd,ID_REMIX+orgMix-1,TRUE);
    CheckDlgButton(hwnd,(orgFreq==11) ? ID_11025 : ID_22050,TRUE);
    #if defined __WIN32__ || defined _WIN32 || defined __NT__
      SendDlgItemMessage(hwnd,ID_BLOCKS,TBM_SETRANGE,TRUE,MAKELONG(3,6));
      SendDlgItemMessage(hwnd,ID_BLOCKS,TBM_SETPOS,TRUE,orgBuffers);
    #else
      SetScrollRange(GetDlgItem(hwnd,ID_BLOCKS),SB_CTL,3,6,FALSE);
      SetScrollPos(GetDlgItem(hwnd,ID_BLOCKS),SB_CTL,orgBuffers,TRUE);
    #endif
    // create font for the test pad
    hfont=CreateFont(48, 0, 0, 0, FW_BOLD, TRUE, 0, 0, DEFAULT_CHARSET,
                     OUT_DEFAULT_PRECIS, CLIP_CHARACTER_PRECIS,
                     PROOF_QUALITY, FF_DONTCARE, "Arial");
    SendDlgItemMessage(hwnd,ID_TESTPAD,WM_SETFONT,(WPARAM)hfont,TRUE);
    // create brushes for custom colours
    hbrushGreen=CreateSolidBrush(RGB(0,128,0));
    // subclass the test pad, for the mouse messages
    lpfnTestPadPrev = (FARPROC)GetWindowLong(GetDlgItem(hwnd,ID_TESTPAD),GWL_WNDPROC);
    SetWindowLong(GetDlgItem(hwnd,ID_TESTPAD),GWL_WNDPROC,(LONG)(FARPROC)TestPadShell);
    // set initial info text
    SetDlgItemText(hwnd,ID_INFO,szInfo[0]);
    // Centre dialog
    CentreWindow(hwnd);
    return TRUE;

  case WM_COMMAND:
    switch (LOWORD(wParam)) {
    case IDCANCEL:
      // check whether the settings changed
      if (hMixSession==NULL)
        hMixSession=WaveMixInit();
      curMix=WaveMixGetProfileInt("Remix",1);
      curFreq=WaveMixGetProfileInt("SamplesPerSec",11);
      curBuffers=WaveMixGetProfileInt("WaveBlocks",3);
      if (curMix!=orgMix || curFreq!=orgFreq || curBuffers!=orgBuffers) {
        int result=MessageBox(hwnd,"Keep the current settings?","MixConf",MB_YESNO|MB_ICONQUESTION);
        if (result==IDNO) {
          // reset the fields
          WaveMixSetProfileInt("Remix",orgMix);
          WaveMixSetProfileInt("SamplesPerSec",orgFreq);
          WaveMixSetProfileInt("WaveBlocks",orgBuffers);
        } /* if */
      } /* if */
      CloseWaveMix();
      EndDialog(hwnd, TRUE);
      return (TRUE);
    case ID_REMIX:
      SetDlgItemText(hwnd,ID_INFO,szInfo[1]);
      // save this setting immediately
      if (hMixSession==NULL)
        hMixSession=WaveMixInit();
      WaveMixSetProfileInt("Remix",1);
      CloseWaveMix();
      return TRUE;
    case ID_QUEUE:
      SetDlgItemText(hwnd,ID_INFO,szInfo[2]);
      // save this setting immediately
      if (hMixSession==NULL)
        hMixSession=WaveMixInit();
      WaveMixSetProfileInt("Remix",2);
      CloseWaveMix();
      return TRUE;
    case ID_NORESET:
      SetDlgItemText(hwnd,ID_INFO,szInfo[3]);
      // save this setting immediately
      if (hMixSession==NULL)
        hMixSession=WaveMixInit();
      WaveMixSetProfileInt("Remix",3);
      CloseWaveMix();
      return TRUE;
    case ID_11025:
      SetDlgItemText(hwnd,ID_INFO,szInfo[4]);
      // save this setting immediately
      if (hMixSession==NULL)
        hMixSession=WaveMixInit();
      WaveMixSetProfileInt("SamplesPerSec",11);
      CloseWaveMix();
      return TRUE;
    case ID_22050:
      SetDlgItemText(hwnd,ID_INFO,szInfo[5]);
      // save this setting immediately
      if (hMixSession==NULL)
        hMixSession=WaveMixInit();
      WaveMixSetProfileInt("SamplesPerSec",22);
      CloseWaveMix();
      return TRUE;
    case ID_CONTINUOUS:
      PlayBackgr=IsDlgButtonChecked(hwnd,ID_CONTINUOUS);
      return TRUE;
    } /* switch */
    break;

#if defined __WIN32__ || defined _WIN32 || defined __NT__
  case WM_CTLCOLORSTATIC:
#else
  case WM_CTLCOLOR:
#endif
    if ((HWND)lParam==GetDlgItem(hwnd,ID_TESTPAD)) {
      SetTextColor((HDC)wParam, RGB(128,0,0));
      SetBkColor((HDC)wParam, RGB(0,128,0));
      return (BOOL)hbrushGreen;
    } else if ((HWND)lParam==GetDlgItem(hwnd,ID_INFO)) {
      SetTextColor((HDC)wParam, RGB(0,0,0));
      SetBkColor((HDC)wParam, RGB(255,255,255));
      return (BOOL)GetStockObject(WHITE_BRUSH);
    } /* if */
    return FALSE;

  case WM_DESTROY:
    DeleteObject(hbrushGreen);
    break;

  case WM_HSCROLL:
    SetDlgItemText(hwnd,ID_INFO,szInfo[6]);
    #if defined __WIN32__ || defined _WIN32 || defined __NT__
      pos=SendDlgItemMessage(hwnd,ID_BLOCKS,TBM_GETPOS,0,0L);
    #else
      pos=GetScrollPos(GetDlgItem(hwnd,ID_BLOCKS),SB_CTL);
      switch (wParam) {
      case SB_LINEDOWN:
        if (pos<6)
          pos++;
        break;
      case SB_LINEUP:
        if (pos>3)
          pos--;
        break;
      case SB_THUMBPOSITION:
        pos=LOWORD(lParam);
        break;
      } /* switch */
      SetScrollPos(GetDlgItem(hwnd,ID_BLOCKS),SB_CTL,pos,TRUE);
    #endif
    // save this setting immediately
    if (hMixSession==NULL)
      hMixSession=WaveMixInit();
    WaveMixSetProfileInt("WaveBlocks",pos);
    CloseWaveMix();
    return TRUE;

  } /* switch */
  return FALSE;
}

#pragma argsused
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
                   LPSTR lpCmdLine, int nCmdShow)
{
  hinst=hInstance;
  #if defined __WIN32__ || defined _WIN32 || defined __NT__
    InitCommonControls();
  #else
    hinst3D = 0;        /* preset */
    {
      WORD sem = SetErrorMode(SEM_NOOPENFILEERRORBOX);
      if ((hinst3D=LoadLibrary("ctl3dv2.dll"))>HINSTANCE_ERROR
          || (hinst3D=LoadLibrary("ctl3d.dll"))>HINSTANCE_ERROR)
      {
        (FARPROC)Ctl3dRegister = GetProcAddress(hinst3D, "Ctl3dRegister");
        (FARPROC)Ctl3dUnregister = GetProcAddress(hinst3D, "Ctl3dUnregister");
        (FARPROC)Ctl3dAutoSubclass = GetProcAddress(hinst3D, "Ctl3dAutoSubclass");
        Ctl3dRegister(hInstance);
        Ctl3dAutoSubclass(hInstance);
      } /* if */
      SetErrorMode(sem);
    }
  #endif

  DialogBox(hInstance,"MixConf",NULL,(DLGPROC)MixConfDlg);

  /* Clean up */
  #if defined __WIN32__ || defined _WIN32 || defined __NT__
  #else
    if (hinst3D != 0) {
      Ctl3dUnregister(hInstance);
      FreeLibrary(hinst3D);
    } /* if */
  #endif

  return 0;
}
